#!/usr/bin/perl -w
use strict;
use English;



sub destructive_merge_mangle {
  my($destination_ref, $source_ref, $comparison, $mangler) = @_;
  # Merges elements of destination into source according to the
  # comparison function.  Assumes that both source and destination
  # are already sorted wrt comparison.

  # mangler is optional, but if provided will be called for each
  # element of the modified destination with the args
  # ($destination_ref, $index, $new_item_p).  You can check for undef
  # on the next or previous index values to see if you're inserting at
  # the end/beginning of the list.

  my @source = @$source_ref;
  my $src_head = shift @source;
  my $current_splice_pos = 0;
  my $dest_items_left = scalar(@$destination_ref);
  my $next_dest = $$destination_ref[$current_splice_pos];
  while($src_head && $dest_items_left) {
    if(&$comparison($next_dest, $src_head) == 1) {
      # i.e. next_dest > src_head
      splice @$destination_ref, $current_splice_pos, 0, $src_head;
      &$mangler($destination_ref, $current_splice_pos, 1) if $mangler;
      $src_head = shift @source;
      $current_splice_pos++;
    } else {
      &$mangler($destination_ref, $current_splice_pos, 0) if $mangler;      
      $current_splice_pos++;
      $next_dest = $$destination_ref[$current_splice_pos];
      $dest_items_left--;
    }
  }
  if($src_head) {
    push @$destination_ref, $src_head, @source;
    if($mangler) {
      my $tail;
      foreach $tail ($src_head, @source) {
        &$mangler($destination_ref, $current_splice_pos, 1) if $mangler;
        $current_splice_pos++;
      }
    }
  } else {
    # Must be some destination_ref items left.
    while ($dest_items_left) {
      &$mangler($destination_ref, $current_splice_pos, 0) if $mangler;
      $current_splice_pos++;
      $dest_items_left--;
    }
  }
}





sub destructive_remove_mangle {
  my($destination_ref, $source_ref, $comparison, $mangler) = @_;

  # Removes elements of destination from source according to the
  # comparison function which is just boolean.  Assumes that source
  # items are in the same order in source_ref as they are in
  # destination_ref

  # mangler is optional, but if provided will be called for each item
  # in the new destination list with the args:

  # ($killed_p, $destination_ref, $index, $old_item).  If killed is true
  # ($killed_p, $destination_ref, $index).  If killed is false

  # if killed is true, index is the delete pos, while if false, it's
  # the new index.  You can check for undef on the next or previous
  # index values to see if you're at the end/beginning of the list.

  my @source = @$source_ref;
  my $src_head = shift @source;
  my $current_splice_pos = 0;
  my $dest_items_left = scalar(@$destination_ref);
  my $next_dest = $$destination_ref[$current_splice_pos];
  while($src_head && $dest_items_left) {
    if(&$comparison($next_dest, $src_head)) {
      # found an item to delete.
      splice @$destination_ref, $current_splice_pos, 1;
      &$mangler(1, $destination_ref, $current_splice_pos, $src_head)
        if $mangler;
      $src_head = shift @source;
    } else {
      &$mangler(0, $destination_ref, $current_splice_pos);      
      $current_splice_pos++;
    }
    $dest_items_left--;
    $next_dest = $$destination_ref[$current_splice_pos];
  }
  if($src_head) {
    print STDERR
      "Warning: source items left after destructive_remove_mangle\n";
  }
  if($mangler && $dest_items_left) {
    while($dest_items_left) {
      &$mangler(0, $destination_ref, $current_splice_pos);
      $current_splice_pos++;
      $dest_items_left--;
    }
  }
}

my @a = (1, 2, 3, 4, 5, 6, 7, 8, 9);
my @b = (2, 4, 6, 8);

#destructive_merge_mangle(\@a, \@b, sub { return $_[0] <=> $_[1]; },
#                         sub { print join(' ', @_) . "\n"; });
#

destructive_remove_mangle(\@a, \@b, sub { return $_[0] == $_[1]; },
                          sub { print join(' ', @_) . "\n"; });



print join(' ', @a) . "\n";







__END__


use Tk;
use Tk::ROText;

my $mw = new MainWindow();
my $list = $mw->ROText(-width => 80,
                       -height => 10,
                       -spacing1 => 1,
                       -takefocus => 0,
                       -exportselection => 0,
                       -wrap => 'none',
                       -relief => 'sunken');
$list->pack(-side => 'top', -fill => 'both', -expand => 1);  

$list->tagConfigure('color2',
                    -background => "green",
                    -foreground => 'black');

$list->insert('end', "foo\n", 'color2');
$list->delete('end - 1 c');

#$list->insert('end', "bar\n");
#$list->insert('end', "foo\n", 'color2');

print "[" . $list->get('end - 1c') . "]\n";

MainLoop();

__END__

require 'Cbblib.pm';

my $db = new Cbblib::Db;
$db->load_file($ARGV[0]);

my $mw = new MainWindow();
my $visible_lines = 20;

my $frame = $mw->Frame(-relief => 'sunken', -borderwidth => 4);
$frame->pack(-side => 'top', -fill => 'both', -expand => 1);
my $list = $frame->Text(-height => $visible_lines * 2, -width => 100,
                        -font => '-schumacher-clean-medium-r-normal--12-*');
$list->pack(-side => 'top', -fill => 'both', -expand => 1);

#$list->tagConfigure('color1', -background => 'wh');
$list->tagConfigure('color2', -background => 'grey78');


my $items = $db->get_txns();
my @items = @$items;

sub txn_to_text {
  my($txn) = @_;
  my $splits = $txn->get_splits_();
  
  my $splits_text = "";
  my $split;
  foreach $split (@$splits) {
    my $dest = $split->get_dest();
    if(ref($dest) eq 'Cbblib::Acct') {
      $dest = '[' . $dest->get_name() . ']';
    } else {
      $dest = $dest->get_name();
    }      

    $splits_text .= 
      sprintf("           %-15.15s %-20.20s %9.2f  %9.2f\n", 
              $dest,
              $split->get_notes(),
              $split->get_debit(),
              $split->get_credit()
              );
  }
  return(sprintf("%5.5s  %-8.8s  %-29.29s  %9.2f  %9.2f  %-1.1s  %9.2f\n",
                 $txn->get_checkno(),
                 $txn->get_date(),
                 $txn->get_desc(),
                 0, 0,
                 $txn->get_status(),
                 0),
         $splits_text);
}

my $color = "";
while ($#items) {
  #if($color eq '') {
  #  $color = 'color2';
  #} else {
  #  $color = '';
  #}
  my $txn = shift(@items);
  my($l1, $l2) = txn_to_text($txn);
  $list->insert('end', $l1);
  $list->insert('end', $l2, 'color2');
}

MainLoop();

__END__

sub add_item {
  my $txn = shift(@items);
  my($l1, $l2) = txn_to_text($txn);
  $list->delete('1.0', '2.0');
  $list->insert('end', $l1, 'color1');
  $list->insert('end', $l2, 'color2');
  $mw->after(2, \&add_item);
}

for($i=0; $i < $visible_lines; $i++) {
  my $txn = shift(@items);
  my($l1, $l2) = txn_to_text($txn);
  $list->insert('end', $l1, 'color1');
  $list->insert('end', $l2, 'color2');
}



sub add_item {
  my $dead = shift @visible_items;
  my $txn = shift(@items);
  my $item = $frame->Text(-height => 1);
  $item->insert('end', txn_to_text($txn));
#  my $item = $frame->Text(-text => txn_to_text($txn));
  $dead->destroy();
  $item->pack(-side => 'top', -fill => 'x');
  push @visible_items, $item;  
  $i++;
  $mw->update();
  $mw->after(1, \&add_item);
}

for($i=0; $i < $visible_lines; $i++) {
  my $txn = shift(@items);
  my $item = $frame->Text(-height => 1);
  $item->insert('end', txn_to_text($txn));
  $item->pack(-side => 'top', -fill => 'x');
  push @visible_items, $item;
}

$mw->after(100, \&add_item);

MainLoop();





