use strict;
use Tk;
use Tk::ROText;

# Hash containing hash refs for all the open Account windows.
my %ledger_wins;

# Each hash ref has the following legitimate values:
#
# acct -- The actual acct (the key).
# tk-win -- top level window
# ledger -- the actual ledger
# line-to-ledger -- ref to array mapping text lines to ledger entries 
# txn-to-lines -- ref to hash mapping txns to [ start-line, count ] pairs
# entry-values -- ref to hash of textvariables for entry fields
# current-txn -- Current transaction being edited (if any) and split_line.
#   Format is [txn, split_line, editing_txn, new_split]
#   where editing_txn indicates that this transaction's being edited in the
#   entry fields and new_split is a boolean indicating that the split line
#   hasn't actually been committed yet.
# callbacks -- ref to hash of callbacks currently registered.
# txns-widget -- widget containing the transactions.

# The idea with split lines is that when the user says "new split
# line" we add a blank split line (only visually) to the currently
# selected transaction and drop them into the edit fields.  Then we
# make sure that any time they select another transaction, the blank
# split line dies.  Any editing they've done to the uncommitted split
# line should be subject to the same warnings that come with any other
# uncommitted edit.

sub desired_ledger_list_size {
  $pref_ledger_list_size =~ /(\d+)x(\d+)/;
  return ($1, $2);
}

sub string_to_acct_or_cat {
  my($string) = @_;
  my $result;

  if($string =~ /^\[(.*)\]$/mo) {
    $result = $main::db->get_accts_by_name($1);
    if($$result[0]) {
      return $$result[0];
    }
  }

  if(!$string) { return $main::db->get_default_sink(); }

  $result = $main::db->get_cats_by_name($string);
  if($$result[0]) {
    return $$result[0];
  } else {
    return undef;
  }
}


sub ledger_get_txn_position {
  my($ledger, $txn) = @_;
  my $result = undef;
  
  # Should use binary search!
  my $pos = 0;
  while(!$result && ($pos <= $#$ledger)) {
    my $item = $$ledger[$pos];
    if($$item[0] == $txn) {
      $result = $pos;
    }
    $pos++;
  }
  return $result;
}

sub ledger_item_to_text {
  my($acct, $item) = @_;
  my $txn = $$item[0];
  my $total = $$item[1];
  my $splits = $txn->get_splits_();
  my $source = $txn->get_source();
  
  my($total_debit, $total_credit) = $txn->totals_wrt($acct);
  my @result = ();

  push @result, 
  sprintf("%5.5s  %-8.8s  %-29.29s  %9.2f  %9.2f  %-1.1s  %9.2f",
          $txn->get_checkno(),
          $txn->get_date(),
          $txn->get_desc(),
          $total_debit, $total_credit,
          $txn->get_status_wrt($acct),
          $total);
  my $split;
  foreach $split (@$splits) {

    my $dest = $split->get_dest();
    next if($source != $acct && $dest != $acct);

    # If this split line is being displayed in the dest rather
    # than the source
    if($acct == $dest) { $dest = $source; }
    
    my $dest_name;
    if(ref($dest) eq 'CBBlib::Acct') {
      $dest_name = '[' . $dest->get_name() . ']';
    } else {
      $dest_name = $dest->get_name();
    }
    
    my $split_line;
    if($txn->get_source() == $acct) {
      $split_line = 
        sprintf("           %-15.15s %-20.20s %9.2f  %9.2f", 
                $dest_name,
                $split->get_notes(),
                $split->get_debit(),
                $split->get_credit());
    } else {
      $split_line = 
        sprintf("           %-15.15s %-20.20s %9.2f  %9.2f", 
                $dest_name,
                $split->get_notes(),
                $split->get_credit(),
                $split->get_debit());
    }
    push @result, $split_line;
  }
  return \@result;
}


sub ledger_win_add_menus {
  my($menubar) = @_;
  
  my %mb; 
  $mb{'file'} = 
    $menubar->Menubutton(-text => 'File', -underline => 1, -takefocus => 0);
  $mb{'edit'} = 
    $menubar->Menubutton(-text => 'Edit', -underline => 1, -takefocus => 0);
  $mb{'extern'} = 
    $menubar->Menubutton(-text => 'External', -underline => 1,
                         -takefocus => 0);
  if($pref_devel) {
    $mb{'devel'} = 
      $menubar->Menubutton(-text => 'Devel', -takefocus => 0);
  }  
  $mb{'help'} =
    $menubar->Menubutton(-text => 'Help', -underline => 0, -takefocus => 0);

  my $help_menu = menubutton_add_help_menu($mb{'help'});
  
  my @mbs = ($mb{'file'}, $mb{'edit'}, $mb{'extern'});
  if($mb{'devel'}) { push @mbs, $mb{'devel'}; }
  my $mb;
  foreach $mb (@mbs) { $mb->pack(-side => 'left', -ipadx => 2); }
  $mb{'help'}->pack(-side => 'right', -ipadx => 1);
  
  return %mb;
}

sub ledger_win_setup_file_menu {
  my($mb, $data) = @_;

  my $win = $$data{'tk-win'};
  my $menu = $mb->Menu(-tearoff => 0);

  $menu->add('command', -label => 'Save Database',
             -underline => 0,
             -command => \&db_save);

  $menu->add('command', -label => 'Save Database As...',
             -underline => 5,
             -command => \&db_save_as);    

  $menu->add('separator');

  $menu->add('command', -label => 'Balance',
             -underline => 0,
             -command => sub { new_balance_win($$data{'acct'}); });
  $menu->entryconfigure(3, -accelerator => '<Meta-b>');
  $win->bind('<Meta-b>', sub { new_balance_win($$data{'acct'}); });
  $win->bind('<Alt-b>', sub { new_balance_win($$data{'acct'}); });


  $menu->add('separator');

  $menu->add('command', -label => 'Close',
             -underline => 0,
             -command => \&unimplemented_window);
  # Close should:
  #   Ask to save if needed
  #   close the win
  #   delete ledger_wins{$$data{'acct'}}
  #   release_ledger()

  $menu->add('separator');

  $menu->add('command', -label => 'Quit',
             -command => \&application_quit);

  $mb->configure(-menu => $menu);
}

sub ledger_win_setup_edit_menu {
  my($mb, $data) = @_;

  my $win = $$data{'tk-win'};
  my $menu = $mb->Menu(-tearoff => 0);

  $menu->add('command', -label => 'Undo', -underline => 0,
             -command => \&unimplemented_window);
  $menu->entryconfigure(0, -accelerator => '<Meta-U>');
  $win->bind('<Meta-u>', \&undoAction);
  $win->bind('<Alt-u>', \&undoAction);

  $menu->add('separator');
  
  $menu->add('command', -label => 'New Transaction',
             -underline => 0,
             -command => sub { ledger_win_edit_new_txn($data); });
  $menu->entryconfigure(2, -accelerator => '<Meta-N>');
  $win->bind('<Meta-n>', sub { ledger_win_edit_new_txn($data); });
  $win->bind('<Alt-n>', sub { ledger_win_edit_new_txn($data); });
  
  $menu->add('command', -label => 'New Split Line',
             -underline => 0,
             -command => sub { ledger_win_add_split_line($data, 'below'); });
  $menu->entryconfigure(3, -accelerator => '<Meta-x>');
  $win->bind('<Meta-x>', sub { ledger_win_add_split_line($data, 'below'); });
  $win->bind('<Alt-x>', sub { ledger_win_add_split_line($data, 'below'); });

  $menu->add('command', -label => 'Clear Transaction',
             -underline => 0,
             -command => sub { ledger_win_clear_txn($data); });
  $menu->entryconfigure(4, -accelerator => '<Meta-k>');
  $win->bind('<Meta-k>', sub { ledger_win_clear_txn($data); });
  $win->bind('<Alt-k>', sub { ledger_win_clear_txn($data); });
  
  $menu->add('command', -label => 'Delete Transaction',
             -underline => 0,
             -command => sub { ledger_win_delete_txn($data); });

  $menu->add('command', -label => 'Commit Transaction',
             -underline => 0,
             -command => sub { ledger_win_finish_edit_txn($data); });
  $menu->entryconfigure(6, -accelerator => '<Meta-d>');
  $win->bind('<Meta-d>', sub { ledger_win_finish_edit_txn($data); });
  $win->bind('<Alt-d>', sub { ledger_win_finish_edit_txn($data); });

  $win->bind('<Meta-s>', \&db_save);
  $win->bind('<Alt-s>', \&db_save);

#     $me add command -label 'Edit Transaction' -underline 0 \
# 	    -command { 
# 	if { [listGetCurTrans] >= 1 } {
# 	    update_entry_area [listGetCurTrans].0
# 	}
#     }
#     $me entryconfig 3 -accelerator <Meta-E>
#     bind all <Meta-e> { 
# 	if { [listGetCurTrans] >= 1 } {
# 	    update_entry_area [listGetCurTrans].0
# 	}
#     }
#     bind all <Alt-e> { 
# 	if { [listGetCurTrans] >= 1 } {
# 	    update_entry_area [listGetCurTrans].0
# 	}
#     }

#     $me add command -label 'Delete Transaction' -underline 0 \
# 	    -command { 
# 	if { "[.trans.list tag nextrange sel 1.0 end ]" != "" } {
# 	    delete_trans [.trans.list tag nextrange sel 1.0 end ]
# 	}
#     }

#     $me add separator

#     $me add command -label 'Open Category Splits ... ' -underline 0 \
# 	    -command { 
# 	cbbWindow.splits
# 	tkwait window .splits
#     }
#     $me entryconfig 6 -accelerator <Meta-S>
#     bind all <Meta-s> { 
# 	cbbWindow.splits
# 	tkwait window .splits
#     }
#     bind all <Alt-s> { 
# 	cbbWindow.splits
# 	tkwait window .splits
#     }

  $mb->configure(-menu => $menu);
}


sub ledger_win_add_menubar {
  my($data) = @_;
  my $parent = $$data{'tk-win'};
  my $menubar = $parent->Frame(-relief => 'raised', -borderwidth => 2);
  $menubar->pack(-fill => 'x');
  my %mbs = ledger_win_add_menus($menubar);
  
  ledger_win_setup_file_menu($mbs{'file'}, $data);
  ledger_win_setup_edit_menu($mbs{'edit'}, $data);

}

#------------------------------------------------------------------------------
# Setup headers
#------------------------------------------------------------------------------

sub ledger_win_add_headers {
  my($parent) = @_;
  
  my $L1 =
    $parent->Label(-font => $pref_fixed_header_font,
                   -text =>
                   sprintf('%5s  %-8s  %-29s  %9s  %9s  %-1.1s  %9s',
                           'Chk #', 'Date', 'Description', 'Debit', 'Credit',
                           '', 'Total'),
                   -padx => 5, -pady => -1,
                   -foreground => $pref_heading_color);
  
  my $L2 =
    $parent->Label(-font => $pref_fixed_header_font,
                   -text => 
                   sprintf('           %-15s %-20s %9s  %9s', 
                           'Destination', 'Notes', 'Debit', 'Credit'),
                   -padx => 4, -pady => -1,
                   -foreground => $pref_heading_color);
  $L1->pack(-side => 'top', -anchor => 'w');
  $L2->pack(-side => 'top', -anchor => 'w');
}



sub setup_ledger_win_entry_area {
  my($data) = @_;

  my $parent = $$data{'tk-win'};

  my %entry_vals;

  $parent->optionAdd('*font' => $pref_fixed_font);
  
  #my $filename = $main::cbb{'image-dir'} . '/mini-exclam.gif';  
  #my $done_img = $parent->Photo("done", -file => $filename); 
  
  #$filename = $main::cbb{'image-dir'} . '/mini-cross.gif';  
  #my $cancel_img = $parent->Photo("cancel", -file => $filename); 

  my $txn_info = $parent->Frame();
  $txn_info->pack(-side => 'top', -fill => 'x');

  my $checknum =
    $txn_info->Entry(-relief => 'sunken', -width => 5,
                     -textvariable => \$entry_vals{'checknum'});
  $$data{'first-entry-field'} = $checknum;

  my $date = $txn_info->Entry(-relief => 'sunken', -width => 10,
                              -textvariable => \$entry_vals{'date'});

  my $desc = $txn_info->Entry(-relief => 'sunken', -width => 40,
                              -textvariable => \$entry_vals{'description'});

  my $status = $txn_info->Entry(-relief => 'sunken', -width => 1,
                                -textvariable => \$entry_vals{'status'});

  my $total_label = $txn_info->Label(-text => '    Totals:');

  my $total_debit = $txn_info->Label(-width => 11, -justify => 'right', 
                                     -text => '0.00');

  my $total_credit = $txn_info->Label(-width => 11, -justify => 'right',
                                      -text => '0.00');
  #my $done = $line1->Button(-image => $done_img, -command => \&done_entering,
  #                          -takefocus => 0);
  #cbb_set_balloon($done, 'Done entering <Enter>');
  
  my $spacer0 = $txn_info->Frame(-width => '.1c');
  $spacer0->pack(-side => 'left');

  $checknum->pack(-side => 'left');
  $date->pack(-side => 'left');
  $desc->pack(-side => 'left');
  
  $total_label->pack(-side => 'left');
  $total_debit->pack(-side => 'left');
  $total_credit->pack(-side => 'left');

  #my $space = $line2->Label(-width => 15);
  #my $comment =
  #  $line2->Entry(-width => 15, -relief => 'sunken',
  #                -textvariable => \$entry_vals{'comment'});
  #my $category =
  #  $line2->Entry(-width => 9, -relief => 'sunken',
  #                -textvariable => \$entry_vals{'category'});

  #my $cancel = $line2->Button(-image => $cancel_img,
  #                            -command => \&clear_entry_area,
  #                            -takefocus => 0);
  #cbb_set_balloon($cancel, 'Cancel changes and start over <Meta-N>');
  
  #$space->pack(-padx => 5.7, -side => 'left', );
  #$comment->pack(-side => 'left');
  #$category->pack(-padx => 7, -side => 'left');
  #$cancel->pack(-side => 'right');
  
  # setup some bindings
  
  # Change the bindtags so the following bindings execute first
  # the break makes sure any other default bindings are skipped.
  #$check->bindtags([$check, 'Entry', $main::mw, 'all']);
  #$date->bindtags([$date, 'Entry', $main::mw, 'all']);
  
  # Don't use "all" bindings on this field so we can control what happens
  # with a <Tab> or <Shift-Tab>
  #$desc->bindtags([$desc, 'Entry', $main::mw]);
  #$category->bindtags([$category, 'Entry', $main::mw]);
  
#   $check->bind('<KeyRelease-plus>', sub {
#     print "FOO\n";
#     $check_value = inc_check($check_value);
#     $check->icursor('end');
#     return;
#   }); 
  
#   $check->bind('<KeyRelease-equal>', sub {
#     $check_value = inc_check($check_value);
#     $check->icursor('end');
#     return;
#   });

#   $check->bind('<KeyRelease-minus>', sub {
#     $check_value = dec_check($check_value);
#     $check->icursor('end');
#     return;
#   });
  
#   $check->bind('<KeyRelease-underscore>', sub {
#     $check_value = dec_check($check_value);
#     $check->icursor('end');
#     return;
#   });

#   $date->bind('<KeyRelease-plus>', sub {
#     $nicedate_value = inc_date($nicedate_value);
#     $date->icursor('end');
#     return;
#   });

#   $date->bind('<KeyRelease-equal>', sub {
#     $nicedate_value = inc_date($nicedate_value);
#     $date->icursor('end');
#     return;
#   });

#   $date->bind('<KeyRelease-minus>', sub {
#     $nicedate_value = dec_date($nicedate_value);
#     $date->icursor('end');
#     return;
#   });

#   $date->bind('<KeyRelease-underscore>', sub {
#     $nicedate_value = dec_date($nicedate_value);
#     $date->icursor('end');
#     return;
#   });


  #my $f = $parent->Frame(-relief => 'groove', -borderwidth => 3);
  #$f->pack(-side => 'top');

  my $split_info = $parent->Frame();
  $split_info->pack(-side => 'top', -fill => 'x');
  
  my $bottom_pad = $parent->Frame(-height => '.1c');
  $bottom_pad->pack(-side => 'top');

  my $spacer_tmp = $split_info->Frame(-width => '.8c');
  $spacer_tmp->pack(-side => 'left');
  
  my $destination =
    $split_info->Entry(-relief => 'sunken', -width => 15,
                       -textvariable => \$entry_vals{'destination'});
  $$data{'first-split-field'} = $destination;

  my $notes =
    $split_info->Entry(-relief => 'sunken', -width => 25,
                       -textvariable => \$entry_vals{'notes'});
  my $debit =
    $split_info->Entry(-relief => 'sunken', -width => 11,
                       -textvariable => \$entry_vals{'split-debit'});
  my $credit =
    $split_info->Entry(-relief => 'sunken', -width => 11,
                       -textvariable => \$entry_vals{'split-credit'});
  
  #my $kill = $splits->Button(-text => 'Kill', -pady => -2);
  #my $cancel = $splits->Button(-text => 'Revert', -pady => -2);

  $destination->pack(-side => 'left');
  $notes->pack(-side => 'left');
  $debit->pack(-side => 'left');
  $credit->pack(-side => 'left');

  my $ref = \%entry_vals;
  $$data{'entry-values'} = $ref;
}


sub ledger_win_clear_entry_area {
  my($data) = @_;

  my $entries = $$data{'entry-values'};

  $$entries{'checknum'} = '';
  $$entries{'date'} = '';
  $$entries{'description'} = '';
  $$entries{'status'} = ' ';
  ledger_win_clear_split_entry_area($data);
}

sub ledger_win_clear_split_entry_area {
  my($data) = @_;
  
  my $entries = $$data{'entry-values'};

  $$entries{'destination'} = ''; 
  $$entries{'notes'} = '';
  $$entries{'split-debit'} = '';
  $$entries{'split-credit'} = '';
}

sub ledger_win_edit_new_txn {
  my($data) = @_;
  
  $$data{'current-txn'} = undef;
  my $entries = $$data{'entry-values'};

  ledger_win_clear_entry_area($data);

  # ??? I imagine that cur-date needs to be a per-window thing.
  # ??? Same for 'date'.
  if($current_date) {
    $$entries{'date'} = $current_date;
  } else {
    $$entries{'date'} = nice_date($pref_date_format);
    $current_date = $$entries{'date'};
  }
  $$data{'date'} = raw_date();
  
  # get internal sdate, if sdate is defined
  if($main::cbb{'sdate'}) {
    $main::cbb{'int-sdate'} = start_date($main::cbb{'sdate'});
  }
  if($pref_debug) {
    print $$entries{'date'} . "\n";
    print $$data{'date'} . "\n";
  }
  $main::cbb{'no-more-mem'} = 0;  
  $$data{'first-entry-field'}->focus();;
}

sub ledger_win_start_edit_txn {
  my($data, $txn, $splitline) = @_;

  my $entries = $$data{'entry-values'};

  $$entries{'checknum'} = $txn->get_checkno();
  $$entries{'date'} = $txn->get_date();
  $$entries{'description'} = $txn->get_desc();
  $$entries{'status'} = $txn->get_status_wrt($$data{'acct'});

  my $splits = $txn->get_splits();
  
  my $dest = $$splits[$splitline]->get_dest();
  if(ref($dest) eq 'CBBlib::Acct') {
    $dest = '[' . $dest->get_name() . ']';
  } else {
    $dest = $dest->get_name();
  }      
  $$entries{'destination'} = $dest; 
  $$entries{'notes'} = $$splits[$splitline]->get_notes();
  $$entries{'split-debit'} = $$splits[$splitline]->get_debit();
  $$entries{'split-credit'} = $$splits[$splitline]->get_credit();
}

sub ledger_win_finish_edit_txn {
  my($data) = @_;
  my $entries = $$data{'entry-values'};
  my $selection  = $$data{'current-txn'};
  my $date = $$entries{'date'};
  my $source = $$data{'acct'};
  my $checknum = $$entries{'checknum'};
  my $desc = $$entries{'description'};
  my $status = $$entries{'status'};
  $status = ' ' if ! $status;

  my $dest = string_to_acct_or_cat($$entries{'destination'});
  if(!defined($dest)) {
    ok_window("Invalid destination.  Try again.");
    return;
  }
  my $notes = $$entries{'notes'};
  my $debit = $$entries{'split-debit'};
  my $credit = $$entries{'split-credit'};
  
  $debit = 0 if !$debit;
  $credit = 0 if !$credit;

  my $new_txn = 0;

  if(!$selection) {
    # Adding new transaction.
    debug("No selection.  Adding new transaction.\n");
    $new_txn = new CBBlib::Txn($date, $source, $checknum, $desc, $status);  
    $selection = [$new_txn, 0, 0, 1];
  }
  
  my($txn, $split_line, $editing_txn, $new_split) = @$selection;  
  $$data{'current-txn'} = [ $txn, $split_line, 0, 0];
  if($new_split) {
    # create split and add it.
    my $split = new CBBlib::Split($dest, $notes, $debit, $credit);
    debug("Adding new split $split to $txn at line $split_line with dest $dest\n");
    $txn->add_split($split, $split_line);
  } else {
    # find split and modify it
    debug("Modifying existing split line.\n");
    my $splits = $txn->get_splits_();
    my $split = $$splits[$split_line];

    $main::db->begin_txn_modifications();
    $split->set_dest($dest);
    print "Notes: $notes\n";
    $split->set_notes($notes);
    $split->set_debit($debit);
    $split->set_credit($credit);
    $main::db->end_txn_modifications();
  } 
  if($new_txn) { 
    debug("Adding new transaction.\n");
    $main::db->add_txns([$txn]);
  } else {
    debug("Modifying existing transaction fields.\n");
    $main::db->begin_txn_modifications();
    $txn->set_date($date);
    $txn->set_checkno($checknum);
    $txn->set_desc($desc);
    $main::db->end_txn_modifications();
    #$txn->set_status($status);
  }
  ledger_win_hilite_current_txn($data);
  ledger_win_clear_entry_area($data);
}

sub ledger_win_txns_added_callback {
  my($db, $added_indices, $affected_accts, $data) = @_;

  map {
    my ($acct, $txns, $indices) = @$_;
    if(@$indices) {
      ledger_win_redraw_from_txn_num($ledger_wins{$acct}, $$indices[0]);
    }
  } values(%$affected_accts);
}

sub ledger_win_txns_modified_callback {
  my($db, $modifications, $affected_accts, $data) = @_;

  map {
    my ($acct, $txns, $indices, $moves) = @$_;
    if(@$indices) {
      my $first_mod = $$indices[0];
      if(@$moves) {
        my $first_move = $$moves[0];
        if($$first_move[0] < $first_mod) {
          $first_mod = $$first_move[0];
        }
      } else {
        print STDERR "NO MOVES!\n";
      }
      print "First mod was $first_mod\n";
      ledger_win_redraw_from_txn_num($ledger_wins{$acct}, $first_mod);
    }
  } values(%$affected_accts);
}

sub ledger_win_txns_removed_callback {
  my($db, $dead_txns, $deleted_db_indices, $ledger_removal_info, $data) = @_;
  my($ledger_info) = $$ledger_removal_info{$$data{'acct'}};

  if($ledger_info) {
    my($sink, $removed_items) = @$ledger_info; 
    my $initial_item = $$removed_items[0];
    my $initial_removed_index = $$initial_item[0];
    ledger_win_redraw_from_txn_num($data, $initial_removed_index);
  }
}

sub ledger_win_redraw_from_txn_num {
  # This should redraw the window from the transaction index given downward.
  my($data, $index) = @_;
  
  my $acct = $$data{'acct'};
  my $list = $$data{'txns-widget'};

  debug("ledger_win_redraw_from_txn_num($data, $index);\n");
  
  # save the current listbox view ...
  my @prev_yview = $list->yview();
  debug('Saved current view: (' . join(", ", @prev_yview) . ")\n");
  
  my $items = $$data{'ledger'};
  if($index >= scalar(@$items)) {
    return;
  }

  my @items = @$items;  # Is this a win or loss?
                        # Also be careful, this can change on-the-fly.

#  print "[" . join("]\n[", @items) . "]\n";
 
  # delete everything from the change forward, then rebuild our list from 
  # there
  
  my $txn_to_lines = $$data{'txn-to-lines'};

  my $delete_start_line = 0;
  if($index) {
    my $relevant_txn = $items[$index - 1];
    $relevant_txn = $$relevant_txn[0];
    my $lines = $$txn_to_lines{$relevant_txn};
    $delete_start_line = $$lines[0] + $$lines[1];
  }

  debug("deleting from:  $delete_start_line.0 to end\n");
  $list->delete("$delete_start_line.0", 'end');
  
  if(! $$data{'line-to-ledger'}) {
    $$data{'line-to-ledger'} = [];
  }
  if(! $$data{'txn-to-lines'}) {
    $$data{'txn-to-lines'} = {};
  }
  my $line_to_ledger = $$data{'line-to-ledger'};
  $txn_to_lines = $$data{'txn-to-lines'};
  
  @$line_to_ledger = ();  # we could gain a little efficiency by only
                          # clearing part of this, but I'm too lazy.
  %$txn_to_lines = ();

  my $line_num = 1;
  my $first_item = 1;
  my $add_to_display = 0; $add_to_display = 1 if($index == 0);

  my $selection = $$data{'current-txn'};
  
  my $current_pos = 0;
  while ($current_pos <= $#items) {
    my $item = $items[$current_pos];
    my $lines = ledger_item_to_text($acct, $item);

    if($selection) {
      my($selected_txn, $selected_split,
         $editing_txn, $new_split) = @$selection;
      if($new_split && ($$item[0] == $selected_txn)) {
        splice @$lines, $selected_split + 1, 0, ('');
      }
    }

    my $count = scalar(@$lines);

    if($current_pos == $index) {
      $add_to_display = 1;
      $first_item = 1;
    }

    if($add_to_display) {
      if($first_item && ($list->index('end') ne '2.0')) {
        # Because Tk is brain-dead about insertions (see "man 3tk text" if
        # you don't believe me).
        if($$item[1] < 0) {          
          $list->insert('end', "\n", 'negcolor2');
        } else {
          $list->insert('end', "\n", 'color2');
        }
      }
      if($$item[1] < 0) {          
        $list->insert('end', $$lines[0] . "\n", 'negcolor1');
        shift @$lines;
        $list->insert('end', join("\n", @$lines) . "\n", 'negcolor2');
      } else {
        $list->insert('end', $$lines[0] . "\n", 'color1');
        shift @$lines;
        $list->insert('end', join("\n", @$lines) . "\n", 'color2');        
      }
    }

    my $i;
    foreach $i ($line_num .. ($line_num + $count)) {
      $$line_to_ledger[$i] = $item;
    }
    
    $$txn_to_lines{$$item[0]} = [$line_num, $count];
    $line_num += $count;
    $first_item = 0;
    $current_pos++;
  }

  ledger_win_hilite_current_txn($data);
  
  # now try to restore the current view ...  
  $list->yview('moveto', $prev_yview[0]);
}



sub ledger_win_hilite_current_txn {
  my($data) = @_;
  
  my $list = $$data{'txns-widget'};
  $list->tagRemove('sel', '1.0', 'end');
  $list->tagRemove('cbbSel', '1.0', 'end');
  $list->tagRemove('cbbSel2', '1.0', 'end');
  
  my $selection = $$data{'current-txn'};
  if($selection) {
    my $txn_to_lines = $$data{'txn-to-lines'};
    my($txn, $split_line, $editing_txn, $new_split) = @$selection;
    my $lines = $$txn_to_lines{$txn};
    if($lines) {
      my $start = $$lines[0];
      my $count = $$lines[1];
      
      $list->tagAdd('cbbSel', "$start.0", $start + $count . ".0");
      $list->tagAdd('cbbSel2',
                    ($start + $split_line + 1) . ".0",
                    ($start + $split_line + 2) . ".0");
    } else {
      # Selection no longer valid
      $$data{'current-txn'} = undef;
    }
  } else {
    # Selection no longer valid
    $$data{'current-txn'} = undef;
  }
}

sub click_to_txn_info {
  my($list, $pos, $acct) = @_;
  my $loc = $list->index($pos);
  $loc =~ /(\d+)\.(\d+)/o;
  my $line = $1;
  my $col = $2;
  if($line < 1) { $line = 1; }

  my $data = $ledger_wins{$acct};  
  my $line_to_ledger = $$data{'line-to-ledger'};
  my $txn_to_lines = $$data{'txn-to-lines'};
  my $ledger = $$line_to_ledger[$line];
  my $txn = $$ledger[0];
  my $lines = $$txn_to_lines{$txn};
  my $split_line = ($line - $$lines[0]) - 1;
  if($split_line == -1) { $split_line = 0; }
  
  return($txn, $split_line, $data);
}
  

sub ledger_win_clicked {
  my($list, $pos, $acct) = @_;
  print "click\n";

  my($clicked_txn, $clicked_split_line, $data) = 
    click_to_txn_info($list, $pos, $acct);
  my $selection  = $$data{'current-txn'};

  if($selection) {
    my($selected_txn, $selected_split_line,
       $editing_txn, $new_split) = @$selection;  
    
    $$data{'current-txn'} = [$clicked_txn, $clicked_split_line, 0, 0];
    
    if($new_split &&
       ($selected_txn != $clicked_txn ||
        $selected_split_line != $clicked_split_line)) {
      # Need check for discarding edited data here...
      my $ledger = $$data{'ledger'};
      my $ledger_index = ledger_get_txn_position($ledger, $selected_txn);
      ledger_win_redraw_from_txn_num($data, $ledger_index);
    }
  } else {
    $$data{'current-txn'} = [$clicked_txn, $clicked_split_line, 0, 0];
  }

  ledger_win_hilite_current_txn($data);
  Tk->break();
}

sub ledger_win_double_clicked {
  # We can't assume that ledger_win_clicked has already been called.
  my($list, $pos, $acct) = @_;
  print "double-click\n";

  my($clicked_txn, $clicked_split_line, $data) = 
    click_to_txn_info($list, $pos, $acct);
  my $selection  = $$data{'current-txn'};

  if($selection) {
    my($selected_txn, $selected_split_line,
       $editing_txn, $new_split) = @$selection;  
    
    $$data{'current-txn'} = [$clicked_txn, $clicked_split_line, 1, 0];  
    
    if($new_split &&
       ($selected_txn != $clicked_txn ||
        $selected_split_line != $clicked_split_line)) {
      # Need check for discarding edited data here...
      my $ledger = $$data{'ledger'};
      my $ledger_index = ledger_get_txn_position($ledger, $selected_txn);
      ledger_win_redraw_from_txn_num($data, $ledger_index);
    }
  } else {
    $$data{'current-txn'} = [$clicked_txn, $clicked_split_line, 1, 0];  
  }

  ledger_win_hilite_current_txn($data);
  ledger_win_start_edit_txn($data, $clicked_txn, $clicked_split_line);

  Tk->break();
}

sub ledger_win_add_txns {
  my($data) = @_;
  my $win = $$data{'tk-win'};
  my $acct = $$data{'acct'};

  my $f = $win->Frame();
  $f->pack(-side => 'top', -fill => 'both', -expand => 1);
  
  my($lw, $lh) = desired_ledger_list_size();
  my $list = $f->ROText(-width => $lw,
                        -height => $lh,
                        -spacing1 => 1,
                        -takefocus => 0,
                        -exportselection => 0,
                        -font => $pref_fixed_font,
                        -wrap => 'none',
                        -relief => 'sunken');
  $$data{'txns-widget'} = $list;
  $list->pack(-side => 'left', -fill => 'both', -expand => 1);  


  if($list->depth() > 1) {
    $list->tagConfigure('color1',
                        -background => $pref_list_line1_color,
                        -foreground => 'black');
    $list->tagConfigure('color2',
                        -background => $pref_list_line2_color,
                        -foreground => 'black');
    $list->tagConfigure('negcolor1',
                        -background => $pref_list_line1_color,
                        -foreground => 'red');
    $list->tagConfigure('negcolor2',
                        -background => $pref_list_line2_color,
                        -foreground => 'red');
    $list->tagConfigure('cbbSel',
                        -background => $pref_highlight_color,
                        -foreground => 'black');
    $list->tagConfigure('cbbSel2',
                        -background => 'yellow',
                        -foreground => 'black');
  } else {
    $list->tagConfigure('color1');
    $list->tagConfigure('color2');
  }



  # Add the scrollbar
  my $scroll = $f->Scrollbar(-command => [ $list => 'yview' ]);
  $list->configure(-yscrollcommand => [$scroll => 'set']);
  $scroll->pack(-side => 'right', -fill => 'y');
  
  $list->tagConfigure('color2', -background => 'grey78');

  ledger_win_redraw_from_txn_num($data, 0);
    
  $list->bind('<Button-1>', [ \&ledger_win_clicked, Ev('@'), $acct]);
  $list->bind('<Double-Button-1>', [ \&ledger_win_double_clicked, Ev('@'),
                                    $acct]);

  my $callbacks = $$data{'callbacks'};

  $$callbacks{'txns-added'} = 
    $main::db->add_txns_added_callback(\&ledger_win_txns_added_callback,
                                       $data);
  $$callbacks{'txns-removed'} = 
      $main::db->add_txns_removed_callback(\&ledger_win_txns_removed_callback,
                                           $data);
  $$callbacks{'txns-modified'} = 
    $main::db->add_txns_modified_callback(\&ledger_win_txns_modified_callback,
                                          $data);
}


sub ledger_win_clear_txn {
  my($data) = @_;
  
  my $selection = $$data{'current-txn'};
  if(!$selection) {
    cbb_ok_window("You cannot clear a transaction if you haven't " .
                  'selected one.');
  }
  my($txn, $split_line, $editing_txn, $new_split) = @$selection;
  debug("Clearing " . join(" ", @$selection) . "\n");
  $txn->clear_wrt($$data{'acct'});
}


sub ledger_win_delete_txn {
  my($data) = @_;
  
  my $selection = $$data{'current-txn'};
  if(!$selection) {
    cbb_ok_window("You cannot delete a transaction if you haven't " .
                  'selected one.');
  }
  my($txn, $split_line, $editing_txn, $new_split) = @$selection;

  print "ledger-win: deleting transaction\n";
  $txn->print(\*STDOUT, "  ");
  print "\n\n";

  $main::db->remove_txns([$txn]);
}

sub ledger_win_add_split_line {
  my($data, $direction) = @_;
  
  my $selection = $$data{'current-txn'};
  my($txn, $split_line, $editing_txn, $new_split) = @$selection;
  if(!$txn) {
    ok_window('You cannot add a split line without ' .
                  'selecting a transaction');
    return;
  }

  if($new_split) {
    ok_window('You are already in the process of adding a split line ' . 
                  'to the current transaction.');
    return;
  }

  # This is a gross hack...
  ledger_win_start_edit_txn($data, $txn, $split_line);

  if($direction eq 'below') {
    $$data{'current-txn'} = [$txn, $split_line + 1, 1, 1];
  } else {
    $$data{'current-txn'} = [$txn, $split_line, 1, 1];
  }
  my $ledger = $$data{'ledger'};
  my $ledger_index = ledger_get_txn_position($ledger, $txn);
  ledger_win_redraw_from_txn_num($data, $ledger_index);
  ledger_win_clear_split_entry_area($data);
  $$data{'first-split-field'}->focus();
}




sub open_or_raise_ledger_win {
  my($acct) = @_;
  
  my $data = $ledger_wins{$acct};
  
  if($$data{'tk-win'} && $$data{'tk-win'}->Exists()) {
    $$data{'tk-win'}->raise();
    return;
  }

  $data = {};
  $$data{'callbacks'} = {};
  $$data{'tk-win'} = $main::mw->Toplevel();
  $$data{'acct'} = $acct;
  $$data{'ledger'} = $acct->acquire_ledger();

#  print STDERR "Ledger:\n";
#  my $ledger = $$data{'ledger'};
#  map {
#    print " [" . $$_[0] . ", " . $$_[1] . "]\n";
#  } @$ledger;

  $ledger_wins{$acct} = $data;  
  
  $$data{'tk-win'}->title("Transactions: " . $acct->get_name());
  $$data{'tk-win'}->iconname("Transactions: " . $acct->get_name());

  ledger_win_add_menubar($data);
  ledger_win_add_headers($$data{'tk-win'});
  ledger_win_add_txns($data);
  setup_ledger_win_entry_area($data);
}


1;
__END__
